@@ -174,7 +174,7 @@ void Package::setFallbackPackage(const KPackage::Package &package)
         return;
     }
 
-    d->fallbackPackage = new Package(package);
+    d->fallbackPackage = std::make_unique<Package>(package);
 }
 
 KPackage::Package Package::fallbackPackage() const
@@ -191,6 +191,12 @@ bool Package::allowExternalPaths() const
     return d->externalPaths;
 }
 
+void Package::setMetadata(const KPluginMetaData &data)
+{
+    Q_ASSERT(data.isValid());
+    d->metadata = data;
+}
+
 void Package::setAllowExternalPaths(bool allow)
 {
     d.detach();
@@ -223,11 +229,13 @@ KPluginMetaData Package::metadata() const
         }
     }
 
+    // Set a dummy KPluginMetaData object, this way we don't try to do the expensive
+    // search for the metadata again if none of the paths have changed
     if (!d->metadata) {
-        d->metadata = new KPluginMetaData();
+        d->metadata = KPluginMetaData();
     }
 
-    return *d->metadata;
+    return d->metadata.value();
 }
 
 QString PackagePrivate::unpack(const QString &filePath)
@@ -348,7 +356,7 @@ QString Package::filePath(const QByteArray &fileType, const QString &filename) c
 
     // Nested loop, but in the medium case resolves to just one iteration
     //     qCDebug(KPACKAGE_LOG) << "prefixes:" << d->contentsPrefixPaths.count() << d->contentsPrefixPaths;
-    for (const QString &contentsPrefix : qAsConst(d->contentsPrefixPaths)) {
+    for (const QString &contentsPrefix : std::as_const(d->contentsPrefixPaths)) {
         QString prefix;
         // We are an installed package
         if (d->tempRoot.isEmpty()) {
@@ -358,7 +366,7 @@ QString Package::filePath(const QByteArray &fileType, const QString &filename) c
             prefix = fileType == "metadata" ? d->tempRoot : (d->tempRoot + contentsPrefix);
         }
 
-        for (const QString &path : qAsConst(paths)) {
+        for (const QString &path : std::as_const(paths)) {
             QString file = prefix + path;
 
             if (!filename.isEmpty()) {
@@ -413,7 +421,7 @@ QStringList Package::entryList(const QByteArray &key) const
 
     // qCDebug(KPACKAGE_LOG) << "going to list" << key;
     QStringList list;
-    for (const QString &prefix : qAsConst(d->contentsPrefixPaths)) {
+    for (const QString &prefix : std::as_const(d->contentsPrefixPaths)) {
         // qCDebug(KPACKAGE_LOG) << "     looking in" << prefix;
         const QStringList paths = it.value().paths;
         for (const QString &path : paths) {
@@ -468,6 +476,7 @@ void Package::setPath(const QString &path)
     // hold onto the old pointer just in case it does not, however!
     QExplicitlySharedDataPointer<PackagePrivate> oldD(d);
     d.detach();
+    d->metadata = std::nullopt;
 
     // without structure we're doomed
     if (!d->structure) {
@@ -527,7 +536,7 @@ void Package::setPath(const QString &path)
     // now we search each path found, caching our previous path to know if
     // anything actually really changed
     const QString previousPath = d->path;
-    for (const QString &p : qAsConst(paths)) {
+    for (const QString &p : std::as_const(paths)) {
         d->checkedValid = false;
         QDir dir(p);
 
@@ -539,7 +548,7 @@ void Package::setPath(const QString &path)
             QResource::registerResource(d->rccPath);
 
             // we need just the plugin name here, never the absolute path
-            dir = QDir(QStringLiteral(":/") + defaultPackageRoot() + path.midRef(path.lastIndexOf(QLatin1Char('/'))));
+            dir = QDir(QStringLiteral(":/") + defaultPackageRoot() + QStringView(path).mid(path.lastIndexOf(QLatin1Char('/'))));
         }
 
         d->path = dir.canonicalPath();
@@ -570,8 +579,7 @@ void Package::setPath(const QString &path)
 
     // .. but something did change, so we get rid of our discovery cache
     d->discoveries.clear();
-    delete d->metadata;
-    d->metadata = nullptr;
+    d->metadata = std::nullopt;
 
     // uh-oh, but we didn't end up with anything valid, so we sadly reset ourselves
     // to futility.
@@ -642,7 +650,7 @@ QByteArray Package::cryptographicHash(QCryptographicHash::Algorithm algorithm) c
         qCWarning(KPACKAGE_LOG) << "no metadata at" << metadataPath;
     }
 
-    for (const QString &prefix : qAsConst(d->contentsPrefixPaths)) {
+    for (const QString &prefix : std::as_const(d->contentsPrefixPaths)) {
         const QString basePath = d->path + prefix;
         QDir dir(basePath);
 
@@ -858,7 +866,6 @@ KJob *Package::uninstall(const QString &packageName, const QString &packageRoot)
 PackagePrivate::PackagePrivate()
     : QSharedData()
     , fallbackPackage(nullptr)
-    , metadata(nullptr)
     , externalPaths(false)
     , valid(false)
     , checkedValid(false)
@@ -870,7 +877,9 @@ PackagePrivate::PackagePrivate(const PackagePrivate &other)
     : QSharedData()
 {
     *this = other;
-    metadata = nullptr;
+    if (other.metadata && other.metadata.value().isValid()) {
+        metadata = other.metadata;
+    }
 }
 
 PackagePrivate::~PackagePrivate()
@@ -886,8 +895,6 @@ PackagePrivate::~PackagePrivate()
         QDir dir(tempRoot);
         dir.removeRecursively();
     }
-    delete metadata;
-    delete fallbackPackage;
 }
 
 PackagePrivate &PackagePrivate::operator=(const PackagePrivate &rhs)
@@ -898,16 +905,18 @@ PackagePrivate &PackagePrivate::operator=(const PackagePrivate &rhs)
 
     structure = rhs.structure;
     if (rhs.fallbackPackage) {
-        fallbackPackage = new Package(*rhs.fallbackPackage);
+        fallbackPackage = std::make_unique<Package>(*rhs.fallbackPackage);
     } else {
         fallbackPackage = nullptr;
     }
+    if (rhs.metadata && rhs.metadata.value().isValid()) {
+        metadata = rhs.metadata;
+    }
     path = rhs.path;
     contentsPrefixPaths = rhs.contentsPrefixPaths;
     contents = rhs.contents;
     mimeTypes = rhs.mimeTypes;
     defaultPackageRoot = rhs.defaultPackageRoot;
-    metadata = nullptr;
     externalPaths = rhs.externalPaths;
     valid = rhs.valid;
     return *this;
@@ -971,20 +980,17 @@ void PackagePrivate::createPackageMetadata(const QString &path)
 {
     const bool isDir = QFileInfo(path).isDir();
 
-    delete metadata;
     if (isDir && QFile::exists(path + QStringLiteral("/metadata.json"))) {
-        metadata = new KPluginMetaData(path + QStringLiteral("/metadata.json"));
+        metadata = KPluginMetaData::fromJsonFile(path + QStringLiteral("/metadata.json"));
     } else if (isDir && QFile::exists(path + QStringLiteral("/metadata.desktop"))) {
-        auto md = KPluginMetaData::fromDesktopFile(path + QStringLiteral("/metadata.desktop"), {QStringLiteral(":/kservicetypes5/kpackage-generic.desktop")});
-        metadata = new KPluginMetaData(md);
+        metadata = KPluginMetaData::fromDesktopFile(path + QStringLiteral("/metadata.desktop"), {QStringLiteral(":/kservicetypes5/kpackage-generic.desktop")});
     } else {
         if (isDir) {
             qCDebug(KPACKAGE_LOG) << "No metadata file in the package, expected it at:" << path;
         } else if (path.endsWith(QLatin1String(".desktop"))) {
-            auto md = KPluginMetaData::fromDesktopFile(path, {QStringLiteral(":/kservicetypes5/kpackage-generic.desktop")});
-            metadata = new KPluginMetaData(md);
+            metadata = KPluginMetaData::fromDesktopFile(path, {QStringLiteral(":/kservicetypes5/kpackage-generic.desktop")});
         } else {
-            metadata = new KPluginMetaData(path);
+            metadata = KPluginMetaData(path);
         }
     }
 }
@@ -992,7 +998,7 @@ void PackagePrivate::createPackageMetadata(const QString &path)
 QString PackagePrivate::fallbackFilePath(const QByteArray &key, const QString &filename) const
 {
     // don't fallback if the package isn't valid and never fallback the metadata file
-    if (qstrcmp(key, "metadata") != 0 && fallbackPackage && fallbackPackage->isValid()) {
+    if (key != "metadata" && fallbackPackage && fallbackPackage->isValid()) {
         return fallbackPackage->filePath(key, filename);
     } else {
         return QString();
@@ -1007,8 +1013,8 @@ bool PackagePrivate::hasCycle(const KPackage::Package &package)
 
     // This is the Floyd cycle detection algorithm
     // http://en.wikipedia.org/wiki/Cycle_detection#Tortoise_and_hare
-    KPackage::Package *slowPackage = const_cast<KPackage::Package *>(&package);
-    KPackage::Package *fastPackage = const_cast<KPackage::Package *>(&package);
+    const KPackage::Package *slowPackage = &package;
+    const KPackage::Package *fastPackage = &package;
 
     while (fastPackage && fastPackage->d->fallbackPackage) {
         // consider two packages the same if they have the same metadata
@@ -1018,8 +1024,8 @@ bool PackagePrivate::hasCycle(const KPackage::Package &package)
             qCWarning(KPACKAGE_LOG) << "Warning: the fallback chain of " << package.metadata().pluginId() << "contains a cyclical dependency.";
             return true;
         }
-        fastPackage = fastPackage->d->fallbackPackage->d->fallbackPackage;
-        slowPackage = slowPackage->d->fallbackPackage;
+        fastPackage = fastPackage->d->fallbackPackage->d->fallbackPackage.get();
+        slowPackage = slowPackage->d->fallbackPackage.get();
     }
     return false;
 }
