@@ -7,22 +7,19 @@
 
 #include "kpackagetool.h"
 
-#include <QDebug>
-#include <QJsonDocument>
-#include <QJsonArray>
-#include <KShell>
-#include <KLocalizedString>
 #include <KAboutData>
-#include <KPluginLoader>
+#include <KLocalizedString>
+#include <KShell>
+#include <QDebug>
 
-#include <kpackage/packagestructure.h>
+#include <KJob>
 #include <kpackage/package.h>
 #include <kpackage/packageloader.h>
-#include <KJob>
+#include <kpackage/packagestructure.h>
+#include <kpackage/private/utils.h>
 
 #include <QCommandLineParser>
 #include <QDir>
-#include <QDirIterator>
 #include <QFileInfo>
 #include <QMap>
 #include <QRegularExpression>
@@ -30,62 +27,24 @@
 #include <QStringList>
 #include <QTimer>
 #include <QUrl>
+#include <QVector>
 #include <QXmlStreamWriter>
 
-#include <iostream>
 #include <iomanip>
+#include <iostream>
 
 #include "options.h"
 
 #include "../kpackage/config-package.h"
-//for the index creation function
+// for the index creation function
+#include "../kpackage/package_export.h"
 #include "../kpackage/private/packagejobthread_p.h"
 
+#include "kpackage_debug.h"
+
 Q_GLOBAL_STATIC_WITH_ARGS(QTextStream, cout, (stdout))
 Q_GLOBAL_STATIC_WITH_ARGS(QTextStream, cerr, (stderr))
 
-static QVector<KPluginMetaData> listPackageTypes()
-{
-    QStringList libraryPaths;
-
-    const QString subDirectory = QStringLiteral("kpackage/packagestructure");
-
-    const auto lstPath = QCoreApplication::libraryPaths();
-    for (const QString &dir : lstPath) {
-        QString d = dir + QDir::separator() + subDirectory;
-        if (!d.endsWith(QDir::separator())) {
-            d += QDir::separator();
-        }
-        libraryPaths << d;
-    }
-
-    QVector<KPluginMetaData> offers;
-    for (const QString &plugindir : qAsConst(libraryPaths)) {
-        const QString &_ixfile = plugindir + s_kpluginindex;
-        QFile indexFile(_ixfile);
-        if (indexFile.exists()) {
-            indexFile.open(QIODevice::ReadOnly);
-            QJsonDocument jdoc = QJsonDocument::fromBinaryData(indexFile.readAll());
-            indexFile.close();
-
-            QJsonArray plugins = jdoc.array();
-            for (QJsonArray::const_iterator iter = plugins.constBegin(); iter != plugins.constEnd(); ++iter) {
-                const QJsonObject obj = iter->toObject();
-                const QString candidate = obj.value(QStringLiteral("FileName")).toString();
-                offers << KPluginMetaData(obj, candidate);
-            }
-        } else {
-            QVector<KPluginMetaData> plugins = KPluginLoader::findPlugins(plugindir);
-            QVectorIterator<KPluginMetaData> iter(plugins);
-            while (iter.hasNext()) {
-                auto md = iter.next();
-                offers << md;
-            }
-        }
-    }
-    return offers;
-}
-
 namespace KPackage
 {
 class PackageToolPrivate
@@ -107,8 +66,8 @@ public:
     QCommandLineParser *parser = nullptr;
 };
 
-PackageTool::PackageTool(int &argc, char **argv, QCommandLineParser *parser) :
-    QCoreApplication(argc, argv)
+PackageTool::PackageTool(int &argc, char **argv, QCommandLineParser *parser)
+    : QCoreApplication(argc, argv)
 {
     d = new PackageToolPrivate;
     d->parser = parser;
@@ -127,8 +86,7 @@ void PackageTool::runMain()
         const QString path = d->parser->value(Options::hash());
         KPackage::Package package(&structure);
         package.setPath(path);
-        const QString hash =
-            QString::fromLocal8Bit(package.cryptographicHash(QCryptographicHash::Sha1));
+        const QString hash = QString::fromLocal8Bit(package.cryptographicHash(QCryptographicHash::Sha1));
         if (hash.isEmpty()) {
             d->coutput(i18n("Failed to generate a Package hash for %1", path));
             exit(9);
@@ -171,9 +129,9 @@ void PackageTool::runMain()
         d->packageFile = d->package;
     }
 
-    if (!d->packageFile.isEmpty() && (!d->parser->isSet(Options::type()) ||
-                                      type.compare(i18nc("package type", "wallpaper"), Qt::CaseInsensitive) == 0 ||
-                                      type.compare(QLatin1String("wallpaper"), Qt::CaseInsensitive) == 0)) {
+    if (!d->packageFile.isEmpty()
+        && (!d->parser->isSet(Options::type()) || type.compare(i18nc("package type", "wallpaper"), Qt::CaseInsensitive) == 0
+            || type.compare(QLatin1String("wallpaper"), Qt::CaseInsensitive) == 0)) {
         // Check type for common plasma packages
         KPackage::Package package(&structure);
         QString serviceType;
@@ -181,8 +139,9 @@ void PackageTool::runMain()
 
         if (package.isValid() && package.metadata().isValid()) {
             serviceType = package.metadata().value(QStringLiteral("X-Plasma-ServiceType"));
-            if (serviceType.isEmpty() && !package.metadata().serviceTypes().isEmpty()) {
-                serviceType = package.metadata().serviceTypes().first();
+            const auto serviceTypes = readKPackageTypes(package.metadata());
+            if (serviceType.isEmpty() && !serviceTypes.isEmpty()) {
+                serviceType = serviceTypes.first();
             }
         }
 
@@ -191,7 +150,7 @@ void PackageTool::runMain()
                 type = QStringLiteral("KPackage/Generic");
             } else {
                 type = serviceType;
-                //qDebug() << "fallthrough type is" << serviceType;
+                // qDebug() << "fallthrough type is" << serviceType;
             }
         }
     }
@@ -227,17 +186,18 @@ void PackageTool::runMain()
         exit(0);
 
     } else if (d->parser->isSet(Options::generateIndex())) {
-        recreateIndex();
+        // TODO KF6 Remove
+        qWarning() << "The indexing feature is removed in KPackage 5.82";
         exit(0);
 
     } else if (d->parser->isSet(Options::removeIndex())) {
-        removeIndex();
+        // TODO KF6 Remove
+        qWarning() << "The indexing feature is removed in KPackage 5.82";
         exit(0);
 
     } else {
         // install, remove or upgrade
         if (!d->installer.isValid()) {
-
             d->installer = KPackage::Package(new KPackage::PackageStructure());
         }
 
@@ -245,7 +205,7 @@ void PackageTool::runMain()
 
         if (d->parser->isSet(Options::remove()) || d->parser->isSet(Options::upgrade())) {
             QString pkgPath;
-            for (const QString &t : qAsConst(d->pluginTypes)) {
+            for (const QString &t : std::as_const(d->pluginTypes)) {
                 KPackage::Package pkg = KPackage::PackageLoader::self()->loadPackage(t);
                 pkg.setPath(d->package);
                 if (pkg.isValid()) {
@@ -272,7 +232,7 @@ void PackageTool::runMain()
             d->installer.setPath(pkgPath);
 
             if (!d->parser->isSet(Options::type())) {
-                const auto lst = d->installer.metadata().serviceTypes();
+                const QStringList lst = readKPackageTypes(d->installer.metadata());
                 for (const QString &st : lst) {
                     if (!d->pluginTypes.contains(st)) {
                         d->pluginTypes << st;
@@ -307,7 +267,9 @@ void PackageTool::runMain()
             if (installed.contains(pluginName)) { // Assume it's a plugin name
                 d->installer.setPath(pluginName);
                 KJob *uninstallJob = d->installer.uninstall(pluginName, d->packageRoot);
+                // clang-format off
                 connect(uninstallJob, SIGNAL(result(KJob*)), SLOT(packageUninstalled(KJob*)));
+                // clang-format on
                 return;
             } else {
                 d->coutput(i18n("Error: Plugin %1 is not installed.", pluginName));
@@ -316,11 +278,15 @@ void PackageTool::runMain()
         }
         if (d->parser->isSet(Options::install())) {
             KJob *installJob = d->installer.install(d->packageFile, d->packageRoot);
+            // clang-format off
             connect(installJob, SIGNAL(result(KJob*)), SLOT(packageInstalled(KJob*)));
+            // clang-format on
             return;
         }
         if (d->package.isEmpty()) {
-            qWarning() << i18nc("No option was given, this is the error message telling the user he needs at least one, do not translate install, remove, upgrade nor list", "One of install, remove, upgrade or list is required.");
+            qWarning() << i18nc(
+                "No option was given, this is the error message telling the user he needs at least one, do not translate install, remove, upgrade nor list",
+                "One of install, remove, upgrade or list is required.");
             exit(6);
         }
     }
@@ -388,7 +354,11 @@ void PackageTool::showPackageInfo(const QString &pluginName)
     exit(0);
 }
 
-bool translateKPluginToAppstream(const QString &tagName, const QString &configField, const QJsonObject &configObject, QXmlStreamWriter& writer, bool canEndWithDot)
+bool translateKPluginToAppstream(const QString &tagName,
+                                 const QString &configField,
+                                 const QJsonObject &configObject,
+                                 QXmlStreamWriter &writer,
+                                 bool canEndWithDot)
 {
     const QRegularExpression rx(QStringLiteral("%1\\[(.*)\\]").arg(configField));
     const QJsonValue native = configObject.value(configField);
@@ -422,13 +392,14 @@ bool translateKPluginToAppstream(const QString &tagName, const QString &configFi
 void PackageTool::showAppstreamInfo(const QString &pluginName)
 {
     KPluginMetaData i;
-    //if the path passed is an absolute path, and a metadata file is found under it, use that metadata file to generate the appstream info.
-    // This can happen in the case an application wanting to support kpackage based extensions includes in the same project both the packagestructure plugin and the packages themselves. In that case at build time the packagestructure plugin wouldn't be installed yet
+    // if the path passed is an absolute path, and a metadata file is found under it, use that metadata file to generate the appstream info.
+    // This can happen in the case an application wanting to support kpackage based extensions includes in the same project both the packagestructure plugin and
+    // the packages themselves. In that case at build time the packagestructure plugin wouldn't be installed yet
 
     if (QFile::exists(pluginName + QStringLiteral("/metadata.desktop"))) {
         i = KPluginMetaData::fromDesktopFile(pluginName + QStringLiteral("/metadata.desktop"), {QStringLiteral(":/kservicetypes5/kpackage-generic.desktop")});
     } else if (QFile::exists(pluginName + QStringLiteral("/metadata.json"))) {
-        i = KPluginMetaData(pluginName + QStringLiteral("/metadata.json"));
+        i = KPluginMetaData::fromJsonFile(pluginName + QStringLiteral("/metadata.json"));
     } else {
         QString type = QStringLiteral("KPackage/Generic");
         if (!d->pluginTypes.contains(type) && !d->pluginTypes.isEmpty()) {
@@ -447,47 +418,33 @@ void PackageTool::showAppstreamInfo(const QString &pluginName)
         i = pkg.metadata();
     }
 
-    KPluginMetaData packageStructureMetaData;
-    {
-        const auto lst = listPackageTypes();
-
-        QString type;
-        if (!i.serviceTypes().isEmpty()) {
-            type = i.serviceTypes().constFirst();
-        } else {
-            type = QStringLiteral("KPackage/Generic");
-        }
-
-        for (const KPluginMetaData &md : lst) {
-            if (md.pluginId() == type) {
-                packageStructureMetaData = md;
-                break;
-            }
-        }
-    }
-
     if (!i.isValid()) {
         *cerr << i18n("Error: Can't find plugin metadata: %1\n", pluginName);
         std::exit(3);
         return;
     }
+    QString parentApp = i.value(QLatin1String("X-KDE-ParentApp"));
 
-    QString parentApp = i.rawData().value(QLatin1String("X-KDE-ParentApp")).toString();
-    if (parentApp.isEmpty()) {
-        parentApp = packageStructureMetaData.rawData().value(QLatin1String("X-KDE-ParentApp")).toString();
-    }
-    const QJsonObject rootObject = i.rawData()[QStringLiteral("KPlugin")].toObject();
-
-    // Be super aggressive in finding a NoDisplay property. It can be a top-level property or
-    // inside the KPlugin object, it also can be either a stringy type or a bool type. Try all
-    // possible scopes and type conversions to find NoDisplay
-    for (const auto& object : { i.rawData(), rootObject }) {
-        if (object.value(QLatin1String("NoDisplay")).toBool(false) ||
-                // Standard value is unprocessed string we'll need to deal with.
-                object.value(QLatin1String("NoDisplay")).toString() == QStringLiteral("true")) {
-            std::exit(0);
+#if KPACKAGE_BUILD_DEPRECATED_SINCE(5, 85)
+    KPluginMetaData packageStructureMetaData;
+    {
+        const QStringList packageFormats = readKPackageTypes(i);
+        if (!packageFormats.isEmpty()) {
+            packageStructureMetaData = structureForKPackageType(packageFormats.first());
         }
     }
+    if (parentApp.isEmpty()) {
+        parentApp = packageStructureMetaData.value(QLatin1String("X-KDE-ParentApp"));
+        if (!parentApp.isEmpty()) {
+            qCDebug(KPACKAGE_LOG) << "Implicitly specifying X-KDE-ParentApp by it's parent structure is deprecated and will"
+                                     "be removed in KF6. Either the value should be explicitly set or the default will be used";
+        }
+    }
+#endif
+
+    if (i.value(QStringLiteral("NoDisplay"), false)) {
+        std::exit(0);
+    }
 
     QXmlStreamAttributes componentAttributes;
     if (!parentApp.isEmpty()) {
@@ -510,8 +467,9 @@ void PackageTool::showAppstreamInfo(const QString &pluginName)
         outputDevice = outputFile.data();
     }
 
-    if (!rootObject.contains(QStringLiteral("Description"))) {
-        *cerr << "Error: description missing, will result in broken appdata field as <sumary/> is mandatory at " << QFileInfo(i.metaDataFileName()).absoluteFilePath();
+    if (i.description().isEmpty()) {
+        *cerr << "Error: description missing, will result in broken appdata field as <summary/> is mandatory at "
+              << QFileInfo(i.metaDataFileName()).absoluteFilePath();
         std::exit(10);
     }
 
@@ -525,6 +483,8 @@ void PackageTool::showAppstreamInfo(const QString &pluginName)
     if (!parentApp.isEmpty()) {
         writer.writeTextElement(QStringLiteral("extends"), parentApp);
     }
+
+    const QJsonObject rootObject = i.rawData()[QStringLiteral("KPlugin")].toObject();
     translateKPluginToAppstream(QStringLiteral("name"), QStringLiteral("Name"), rootObject, writer, false);
     translateKPluginToAppstream(QStringLiteral("summary"), QStringLiteral("Description"), rootObject, writer, false);
     if (!i.website().isEmpty()) {
@@ -545,7 +505,7 @@ void PackageTool::showAppstreamInfo(const QString &pluginName)
     if (!authors.isEmpty()) {
         QStringList authorsText;
         authorsText.reserve(authors.size());
-        for (const auto &author: authors) {
+        for (const auto &author : authors) {
             authorsText += QStringLiteral("%1 <%2>").arg(author.name(), author.emailAddress());
         }
         writer.writeTextElement(QStringLiteral("developer_name"), authorsText.join(QStringLiteral(", ")));
@@ -571,11 +531,12 @@ QString PackageTool::findPackageRoot(const QString &pluginName, const QString &p
     Q_UNUSED(prefix);
     QString packageRoot;
     if (d->parser->isSet(Options::packageRoot()) && d->parser->isSet(Options::global()) && !d->parser->isSet(Options::generateIndex())) {
-        qWarning() << i18nc("The user entered conflicting options packageroot and global, this is the error message telling the user he can use only one", "The packageroot and global options conflict with each other, please select only one.");
+        qWarning() << i18nc("The user entered conflicting options packageroot and global, this is the error message telling the user he can use only one",
+                            "The packageroot and global options conflict with each other, please select only one.");
         ::exit(7);
     } else if (d->parser->isSet(Options::packageRoot())) {
         packageRoot = d->parser->value(Options::packageRoot());
-        //qDebug() << "(set via arg) d->packageRoot is: " << d->packageRoot;
+        // qDebug() << "(set via arg) d->packageRoot is: " << d->packageRoot;
     } else if (d->parser->isSet(Options::global())) {
         auto const paths = QStandardPaths::locateAll(QStandardPaths::GenericDataLocation, d->packageRoot, QStandardPaths::LocateDirectory);
         if (!paths.isEmpty()) {
@@ -591,7 +552,7 @@ void PackageTool::listPackages(const QStringList &types, const QString &path)
 {
     QStringList list = d->packages(types, path);
     list.sort();
-    for (const QString &package : qAsConst(list)) {
+    for (const QString &package : std::as_const(list)) {
         d->coutput(package);
     }
     exit(0);
@@ -620,16 +581,12 @@ void PackageToolPrivate::renderTypeTable(const QMap<QString, QStringList> &plugi
         if (pluginIt.value()[1].length() > pathWidth) {
             pathWidth = pluginIt.value()[1].length();
         }
-
     }
 
-    std::cout << nameHeader.toLocal8Bit().constData() << std::setw(nameWidth - nameHeader.length() + 2) << ' '
-              << pluginHeader.toLocal8Bit().constData() << std::setw(pluginWidth - pluginHeader.length() + 2) << ' '
-              << pathHeader.toLocal8Bit().constData() << std::setw(pathWidth - pathHeader.length() + 2) << ' '
-              << std::endl;
-    std::cout << std::setfill('-') << std::setw(nameWidth) << '-' << "  "
-              << std::setw(pluginWidth) << '-' << "  "
-              << std::setw(pathWidth) << '-' << "  "
+    std::cout << nameHeader.toLocal8Bit().constData() << std::setw(nameWidth - nameHeader.length() + 2) << ' ' << pluginHeader.toLocal8Bit().constData()
+              << std::setw(pluginWidth - pluginHeader.length() + 2) << ' ' << pathHeader.toLocal8Bit().constData()
+              << std::setw(pathWidth - pathHeader.length() + 2) << ' ' << std::endl;
+    std::cout << std::setfill('-') << std::setw(nameWidth) << '-' << "  " << std::setw(pluginWidth) << '-' << "  " << std::setw(pathWidth) << '-' << "  "
               << std::endl;
     std::cout << std::setfill(' ');
 
@@ -648,12 +605,14 @@ void PackageToolPrivate::listTypes()
     coutput(i18n("Built in:"));
 
     QMap<QString, QStringList> builtIns;
-    builtIns.insert(i18n("KPackage/Generic"), QStringList() << QStringLiteral("KPackage/Generic") << QStringLiteral(KPACKAGE_RELATIVE_DATA_INSTALL_DIR "/packages/"));
-    builtIns.insert(i18n("KPackage/GenericQML"), QStringList() << QStringLiteral("KPackage/GenericQML") << QStringLiteral(KPACKAGE_RELATIVE_DATA_INSTALL_DIR "/genericqml/"));
+    builtIns.insert(i18n("KPackage/Generic"),
+                    QStringList() << QStringLiteral("KPackage/Generic") << QStringLiteral(KPACKAGE_RELATIVE_DATA_INSTALL_DIR "/packages/"));
+    builtIns.insert(i18n("KPackage/GenericQML"),
+                    QStringList() << QStringLiteral("KPackage/GenericQML") << QStringLiteral(KPACKAGE_RELATIVE_DATA_INSTALL_DIR "/genericqml/"));
 
     renderTypeTable(builtIns);
 
-    const QVector<KPluginMetaData> offers = listPackageTypes();
+    const QVector<KPluginMetaData> offers = KPluginMetaData::findPlugins(QStringLiteral("kpackage/packagestructure"));
 
     if (!offers.isEmpty()) {
         std::cout << std::endl;
@@ -672,89 +631,6 @@ void PackageToolPrivate::listTypes()
     }
 }
 
-void PackageTool::recreateIndex()
-{
-    d->packageRoot = findPackageRoot(d->package, d->packageRoot);
-
-    if (!QDir::isAbsolutePath(d->packageRoot)) {
-        if (d->parser->isSet(Options::global())) {
-
-            // Find package roots
-            QStringList packageRoots;
-            const QVector<KPluginMetaData> offers = listPackageTypes();
-            if (!offers.isEmpty()) {
-                for (const KPluginMetaData &info : offers) {
-                    KPackage::Package pkg = KPackage::PackageLoader::self()->loadPackage(info.pluginId());
-                    const auto proot = pkg.defaultPackageRoot();
-                    if (!proot.isEmpty()) {
-                        packageRoots << pkg.defaultPackageRoot();
-                    }
-                }
-            }
-            const auto lstPath = QStandardPaths::locateAll(QStandardPaths::GenericDataLocation, d->packageRoot, QStandardPaths::LocateDirectory);
-            for (auto const &p : lstPath) {
-                for (const auto &_proot: qAsConst(packageRoots)) {
-                    QFileInfo packageDirInfo(QStringLiteral("%1%2").arg(p, _proot));
-                    if (!packageDirInfo.isWritable()) {
-                        continue;
-                    }
-                    const QString packagedir = packageDirInfo.absoluteFilePath();
-                    if (KPackage::indexDirectory(packagedir, s_kpluginindex)) {
-                        d->coutput(i18n("Generating %1%2", packagedir, s_kpluginindex));
-                    } else {
-                        d->cerror(i18n("Didn't write %1%2", packagedir, s_kpluginindex));
-                    }
-                }
-            }
-            return;
-        } else {
-            d->packageRoot = QStandardPaths::writableLocation(QStandardPaths::GenericDataLocation) + QLatin1Char('/') + d->packageRoot;
-        }
-    }
-
-    if (KPackage::indexDirectory(d->packageRoot, s_kpluginindex)) {
-        d->coutput(i18n("Generating %1/%2", d->packageRoot, s_kpluginindex));
-    } else {
-        d->cerror(i18n("Cannot write %1/%2", d->packageRoot, s_kpluginindex));
-    }
-}
-
-void PackageTool::removeIndex()
-{
-    d->packageRoot = findPackageRoot(d->package, d->packageRoot);
-
-    if (!QDir::isAbsolutePath(d->packageRoot)) {
-        if (d->parser->isSet(Options::global())) {
-            const auto lstEntries = QStandardPaths::locateAll(QStandardPaths::GenericDataLocation, d->packageRoot, QStandardPaths::LocateDirectory);
-            for (auto const &p : lstEntries) {
-                QDirIterator it(p, QStringList(s_kpluginindex), QDir::Files | QDir::Writable | QDir::NoDotAndDotDot, QDirIterator::Subdirectories);
-                qDebug() << "IX index remove" << p;
-                while (it.hasNext()) {
-                    it.next();
-                    const QString path = it.fileInfo().absoluteFilePath();
-                    QFile file(path);
-                    if (!file.remove()) {
-                        d->cerror(i18n("Could not remove index file %1", file.fileName()));
-                    } else {
-                        d->coutput(i18n("Removed %1", file.fileName()));
-                    }
-                }
-            }
-            return;
-        } else {
-            d->packageRoot = QStandardPaths::writableLocation(QStandardPaths::GenericDataLocation) + QLatin1Char('/') + d->packageRoot;
-        }
-    }
-    QFile file(d->packageRoot + s_kpluginindex);
-    if (file.exists()) {
-        if (!file.remove()) {
-            d->cerror(i18n("Could not remove index file %1", file.fileName()));
-        } else {
-            d->coutput(i18n("Removed %1", file.fileName()));
-        }
-    }
-}
-
 void PackageTool::packageInstalled(KJob *job)
 {
     bool success = (job->error() == KJob::NoError);
@@ -780,7 +656,9 @@ void PackageTool::packageUninstalled(KJob *job)
         if (d->parser->isSet(Options::upgrade())) {
             d->coutput(i18n("Upgrading package from file: %1", d->packageFile));
             KJob *installJob = d->installer.install(d->packageFile, d->packageRoot);
+            // clang-format off
             connect(installJob, SIGNAL(result(KJob*)), SLOT(packageInstalled(KJob*)));
+            // clang-format on
             return;
         }
         d->coutput(i18n("Successfully uninstalled %1", d->packageFile));
